(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Check_CRP
imports Check_TP 
begin


subsection {*Proving confluence of TRSs*}

datatype ('f, 'l, 'v) cr_proof =
  SN_WCR "('f, 'l) lab join_info" "('f, 'l, 'v) trs_proof" 
  | Weakly_Orthogonal

fun
  check_cr_proof ::
    "bool \<Rightarrow> shows \<Rightarrow>
     ('tp, ('f::{show,linorder}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> (('f,label_type)lab,string) rules \<Rightarrow>
     ('f, label_type, string) cr_proof \<Rightarrow>
     shows check" 
where
  "check_cr_proof a i I J R (SN_WCR joins_i prf) = debug (i []) ''SN_WCR'' (do {
         let tp = tp_ops.mk I False [] R [];
         check_trs_proof a (i +@+ shows_string ''.1'') I J tp prf
           <+? (\<lambda>s. i +@+ '': error below strong normalization + wcr'' +#+ shows_nl
             +@+ indent s);
         check_critical_pairs R (critical_pairs_impl R R) joins_i
           <+? (\<lambda>s. i +@+ shows ''error when proving local confluence of '' +@+ shows_tp I tp +@+ shows_nl +@+ indent s)
   })" 
| "check_cr_proof a i I J R Weakly_Orthogonal = debug (i []) ''Weakly Orthogonal'' ( do {
           check_left_linear_trs R;
           check_allm (\<lambda> (l,r). check (is_Fun l) (shows ''the TRS has variables as left-hand sides'')) R;
           check_allm (\<lambda> (b,s,t). do {
              check (s = t) (shows ''there is a non-trivial critical pairs: '' +@+ shows_term s +@+ shows '' <- . -> ''
                +@+ shows_term t)
            }) (critical_pairs_impl R R) 
         }  <+? (\<lambda>s. i +@+ shows '': error in checking weakly orthogonal system for the TRS'' +@+ shows_nl +@+ shows_trs R
             +@+ indent s))"


fun cr_assms :: "bool \<Rightarrow> ('f, 'l, 'v)cr_proof \<Rightarrow> ('f, 'l, 'v) prob list" where
  "cr_assms a (SN_WCR _ p) = assms a p"
| "cr_assms a Weakly_Orthogonal = []"

lemma check_cr_proof_with_assms_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and fin: "\<forall>p\<in>set (cr_assms a prf). holds p"
    and ok: "isOK (check_cr_proof a i I J R prf)"
  shows "CR (rstep (set R))"
proof -
  interpret tp_spec I by fact
  from ok fin show ?thesis
  proof (induct "prf" arbitrary: i)
    case (SN_WCR joins_i prof i)
    from SN_WCR(1)
    have cp: "isOK(check_critical_pairs R (critical_pairs_impl R R) joins_i)"
      and ok: "isOK(check_trs_proof a (i +@+ shows_string ''.1'') I J (mk False [] R []) prof)" by (auto simp: Let_def)
    from SN_WCR(2) have a: "\<forall> a \<in> set (assms a prof). holds a" by auto
    from check_critical_pairs[OF cp] have WCR: "WCR (rstep (set R))" .
    from check_trs_proof_with_assms_sound[OF I J a ok] have SN: "SN (rstep (set R))" by simp
    from Newman[OF SN WCR] show ?case .
  next
    case Weakly_Orthogonal
    note ok = Weakly_Orthogonal(1)[simplified]
    from ok have ll: "left_linear_trs (set R)" by simp
    from ok have wf: "\<And> l r. (l,r) \<in> set R \<Longrightarrow> is_Fun l" by auto
    from ok have triv: "\<And> b s t. (b,s,t) \<in> critical_pairs (set R) (set R) \<Longrightarrow> s = t"
      by force
    show ?case 
      by (rule weakly_orthogonal_rstep_CR[OF ll triv wf])
  qed
qed

lemma cr_assms_False[simp]: "cr_assms False prf = []"
  by (induct False "prf" rule: cr_assms.induct) simp_all

lemma check_cr_proof_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
  and ok: "isOK (check_cr_proof False i I J R prf)"
  shows "CR (rstep (set R))"
  by (rule check_cr_proof_with_assms_sound[OF I J _ ok], simp)

datatype ('f, 'l, 'v) ncr_proof =
  SN_NWCR "('f, 'l, 'v) trs_proof" 
  | Non_Join "(('f,'l)lab,'v)term" "(('f,'l)lab,'v)rseq" "(('f,'l)lab,'v)rseq" "(('f,'l)lab,'v) non_join_info"

fun
  check_ncr_proof ::
    "bool \<Rightarrow> shows \<Rightarrow>
     ('tp, ('f::{show,linorder}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> (('f,label_type)lab,string) rules \<Rightarrow>
     ('f, label_type, string) ncr_proof \<Rightarrow>
     shows check" 
where
  "check_ncr_proof a i I J R (SN_NWCR prf) = debug (i []) ''SN_NWCR'' (do {
         let tp = tp_ops.mk I False [] R [];
         check_trs_proof a (i +@+ shows_string ''.1'') I J tp prf
           <+? (\<lambda>s. i +@+ '': error below strong normalization + wcr'' +#+ shows_nl
             +@+ indent s);
         check (\<not> isOK(check_critical_pairs_NF R (critical_pairs_impl R R))) (shows ''all critical pairs are joinable'')
           <+? (\<lambda>s. i +@+ shows ''error when disproving local confluence of '' +@+ shows_tp I tp +@+ shows_nl +@+ indent s)
   })" 
|   "check_ncr_proof a i I J R (Non_Join s seq1 seq2 prf) = debug (i []) ''Non_Join'' (do {
         check_non_join R s seq1 seq2 prf
           <+? (\<lambda>s. i +@+ shows ''error when disproving CR of '' +@+ shows_trs R +@+ shows_nl +@+ indent s)
   })" 


fun ncr_assms :: "bool \<Rightarrow> ('f, 'l, 'v)ncr_proof \<Rightarrow> ('f, 'l, 'v) prob list" where
  "ncr_assms a (SN_NWCR p) = assms a p"
| "ncr_assms a _  = []"

lemma check_ncr_proof_with_assms_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and fin: "\<forall>p\<in>set (ncr_assms a prf). holds p"
    and ok: "isOK (check_ncr_proof a i I J R prf)"
  shows "\<not> CR (rstep (set R))"
proof -
  interpret tp_spec I by fact
  from ok fin show ?thesis
  proof (induct "prf" arbitrary: i)
    case (SN_NWCR prof i)
    from SN_NWCR(2) have a: "\<forall> a \<in> set (assms a prof). holds a" by auto
    from SN_NWCR(1)
    have cp: "\<not> isOK(check_critical_pairs_NF R (critical_pairs_impl R R))"
      and ok: "isOK(check_trs_proof a (i +@+ shows_string ''.1'') I J (mk False [] R []) prof)" by (auto simp: Let_def)
    from check_trs_proof_with_assms_sound[OF I J a ok] have SN: "SN (rstep (set R))" by simp
    from check_critical_pairs_NF_SN[OF SN] cp show ?case by simp
  next
    case (Non_Join s seq1 seq2 prof)
    hence ok: "isOK (check_non_join R s seq1 seq2 prof)" by simp
    from check_non_join_sound[OF ok] show ?case .
  qed
qed

lemma ncr_assms_False[simp]: "ncr_assms False prf = []"
  by (induct False "prf" rule: ncr_assms.induct) simp_all

lemma check_ncr_proof_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
  and ok: "isOK (check_ncr_proof False i I J R prf)"
  shows "\<not> CR (rstep (set R))"
  by (rule check_ncr_proof_with_assms_sound[OF I J _ ok], simp)

definition default_grd_fun :: "((string,'l)lab,string)term \<Rightarrow> ((string,'l)lab,string)term \<Rightarrow> ((string,'l)lab,string)subst"
  where "default_grd_fun s t \<equiv> let 
           F = funs_rule_list (s,t);
           m = fold (\<lambda> f m. case f of Sharp (UnLab g) \<Rightarrow> max (length g) m | _ \<Rightarrow> m) F 0;
           suffix = replicate (Suc m) (CHR ''a'')
         in (\<lambda> x. Fun (Sharp (UnLab (x @ suffix))) [])"

abbreviation Tcap_Not_Unif :: "((string,'l)lab,string)non_join_info" where "Tcap_Not_Unif \<equiv> Tcap_Non_Unif default_grd_fun"



end