(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory DP_Transformation_Impl
imports
  Linear_Orders
  DP_Transformation
  QDP_Framework_Impl 
  Misc (* afp entry from collection entry which defines order on pairs *)
       (* should be replaced by  "~~/src/HOL/Library/Product_ord" if Misc removes product ord def *)
begin


definition
  DP_list :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f \<times> nat)list \<Rightarrow> ('f, 'v) rules"
where
  "DP_list shp R d_list \<equiv>
    concat (map (\<lambda>lr.
      let l = fst lr; s = sharp_term shp l in
      map (\<lambda>u. (s, sharp_term shp u)) (filter
        (\<lambda>u. \<not> l \<rhd> u \<and> is_Fun u \<and> (the (root u)) \<in> set d_list) (supteq_list (snd lr)))) R)"

lemma DP_list[simp]:
  assumes d_list: "d_list = defs_list R"
  shows "set (DP_list shp R d_list) = DP shp (set R)" (is "?L = ?R")
proof -
  from defs_list_sound[of R] d_list
    have d_list: "\<And>fn. defined (set R) fn = (fn \<in> set d_list)" by auto
  show ?thesis
  proof
    show "?L \<subseteq> ?R"
    proof
      fix s t assume "(s, t) \<in> ?L" 
      then obtain l r u where "(l, r) \<in> set R" and "r \<unrhd> u" and "\<not> l \<rhd> u"
        and "s = sharp_term shp l" and "t = sharp_term shp u"
        and "is_Fun u" and "defined (set R) (the (root u))"
          unfolding d_list unfolding DP_list_def Let_def by (auto simp: d_list)
      thus "(s, t) \<in> DP shp (set R)" unfolding DP_def by (cases u) auto      
    qed
  next
    show "?R \<subseteq> ?L"
    proof
      fix s t assume "(s, t) \<in> ?R"
      then obtain lr u where "lr \<in> set R" and "snd lr \<unrhd> u" and "\<not> fst lr \<rhd> u"
        and "s = sharp_term shp (fst lr)" and "t = sharp_term shp u"
        and "is_Fun u" and "defined (set R) (the (root u))" unfolding DP_def by auto
      thus "(s, t) \<in> ?L" unfolding DP_list_def d_list[symmetric] Let_def by auto
    qed
  qed
qed

definition applicable_rule_impl :: "(('f,'v)term \<Rightarrow> bool) \<Rightarrow> ('f,'v)rule \<Rightarrow> bool"
  where "applicable_rule_impl Q lr \<equiv> \<forall>t\<in>set (args (fst lr)). Q t"

context tp_spec
begin 
lemma applicable_rule_impl[simp]:
  "applicable_rule_impl (\<lambda>t. t \<in> NF_terms (set (Q tp))) = applicable_rule (set (Q tp))"
  by (rule ext)
     (simp_all add: applicable_rule_impl_def applicable_rule_def NF_rstep_supt_args_conv)
end

(* COMMENT: we silently handle relative rules as strict rules in this step.
  Advantage: other choices like putting Decreasing rules in relative part can also silently be done, and no explict step: Rw_into_R *) 
definition
  dependency_pairs_tt ::
    "('tp, 'f::{show,linorder}, 'v::{show,linorder}) tp_ops \<Rightarrow>
    ('dpp, 'f, 'v) dpp_ops \<Rightarrow> 'tp \<Rightarrow> ('f \<Rightarrow> 'f) \<Rightarrow>  ('f, 'v) rules \<Rightarrow> 'dpp result"
where
  "dependency_pairs_tt I J tp shp P \<equiv> 
  let R = tp_ops.rules I tp;
      Q = tp_ops.Q I tp;
      iQ = tp_ops.is_QNF I tp;
      U = filter (applicable_rule_impl iQ) R
  in
  check_return (do {
     check_wf_trs U;
     check_allm (\<lambda> q. check_no_var q) Q;
     let Qr = map (\<lambda> q. case q of Fun f ss \<Rightarrow> (f,length ss)) Q;
     let D = defs_list U;
     check_allm (\<lambda> (f,n). check ((shp f,n) \<notin> set D) (shows_string ''sharping '' +@+ shows f +@+ shows_string '' yields the defined symbol '' +@+ shows (shp f))) D;
     check_allm (\<lambda> (f,n). check ((shp f,n) \<notin> set Qr) (shows_string ''sharping '' +@+ shows f +@+ shows_string '' yields the symbol '' +@+ shows (shp f) +@+ shows_string '' which is a root of Q'')) D;
     check_subseteq_sorted (DP_list shp U D) P 
       <+? (\<lambda>dp. ''the DP '' +#+ shows_rule dp
         +@+ '' does not appear in the DP problem'' +#+ shows_nl)
   } <+? (\<lambda>s. ''the DP-transformation is not applied correctly.'' +#+ shows_nl +@+ s))
   (dpp_ops.mk J True True P [] Q [] R)"

lemma dependency_pairs_tt:
  assumes I: "tp_spec I"
  and J: "dpp_spec J"
  and tt: "dependency_pairs_tt I J tp shp P = return d"
  and fin: "finite_dpp (dpp_ops.dpp J d)"
  shows "SN_qrel (tp_ops.qreltrs I tp)"
proof -
  interpret tp_spec I by fact
  interpret spec: dpp_spec J by fact
  let ?R = "set (R tp)"
  let ?Rw = "set (Rw tp)"
  let ?Q = "set (Q tp)"
  let ?Qr = "map (\<lambda>q. case q of Fun f ss \<Rightarrow> (f, length ss)) (Q tp)"
  let ?nfs = "NFS tp"
  let ?B = "set (R tp) \<union> set (Rw tp)"
  let ?u = "{x. (x \<in> set (R tp) \<or> x \<in> set (Rw tp)) \<and> applicable_rule ?Q x}"
  let ?U = "applicable_rules ?Q ?B"
  have u: "?u = ?U" by (auto simp: applicable_rules_def)
  note tt = tt[unfolded dependency_pairs_tt_def Let_def, simplified]
  note ttu = tt[unfolded u]
  from tt have idpp: "d = dpp_ops.mk J True True P [] (Q tp) [] (rules tp)" by auto
  from ttu have P: "DP shp ?U \<subseteq> set P" by (simp add: u)
  have wf: "wf_qtrs ?Q ?B" unfolding wf_qtrs_def wwf_qtrs_def
    using tt[unfolded wf_trs_def] by auto
  hence wwf: "wwf_qtrs ?Q ?B" unfolding wf_qtrs_def by auto
  note switch = wwf_qtrs_imp_nfs_switch[OF wwf]
  have SN: "SN (qrstep True ?Q ?B)"
  proof (rule ccontr)
    assume nSN: "\<not> ?thesis"
    have easy: "\<And> p. \<lbrakk> \<And> f n. p f n\<rbrakk> \<Longrightarrow> True = (\<forall> f n. p f n)" by blast
    have "\<exists> s t \<sigma>. min_ichain (initial_dpp shp True True ?Q ?B) s t \<sigma>"
    proof (rule not_SN_imp_min_ichain[OF wf nSN _ easy])
      fix f n
      show "defined (applicable_rules ?Q ?B) (f,n) \<longrightarrow> 
      \<not> defined (applicable_rules ?Q ?B) (shp f, n)" using ttu by auto
    next
      fix f n and ss :: "('b, 'c) term list"
      assume d: "defined (applicable_rules ?Q ?B) (f,n)" "length ss = n"
      with ttu have not: "(shp f,n) \<notin> set ?Qr" by auto
      from d have n: "n = length ss" by auto
      show "Fun (shp f) ss \<notin> ?Q" using not unfolding n by force
    qed
    then obtain s t \<sigma> where mic: "min_ichain (initial_dpp shp True True ?Q ?B) s t \<sigma>" by blast
    from min_ichain_mono[OF this[unfolded initial_dpp.simps] P _ refl subset_refl refl] P
    have min: "min_ichain (True, True, set P,{},?Q,{},?B) s t \<sigma>" by auto
    with fin[unfolded idpp dpp_spec.mk_sound[OF J]] show False unfolding finite_dpp_def by auto
  qed
  hence "SN (qrstep ?nfs ?Q ?B)" using switch[of ?nfs True] by simp
  hence SN: "SN_qrel (?nfs,?Q,?B,{})" by simp
  show ?thesis unfolding qreltrs_sound
    by (rule SN_qrel_mono[OF subset_refl _ _ SN]) simp_all
qed


end
