(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Matchbounds
imports
  RPO
  Term_Multiset
  Trs_Impl
begin

locale enrichment = 
  fixes R :: "('f,'v)trs"
    and R' :: "('g,'v)trs"
    and lift :: "('f,'v)term \<Rightarrow> ('g,'v)term"
    and base :: "('g,'v)term \<Rightarrow> ('f,'v)term"
 assumes base_lift: "\<And> s :: ('f,'v)term. base (lift s) = s"
   and lift: "\<And> s t s'. (s,t) \<in> rstep R \<Longrightarrow> s = base s' \<Longrightarrow> \<exists> t'. t = base t' \<and> (s',t') \<in> rstep R'"
begin


lemma enrich2: 
  "{t | t s. s \<in> L \<and> (s,t) \<in> (rstep R)^*} \<subseteq> base ` {t'. \<exists> s'. s' \<in> lift ` L \<and> (s',t') \<in> (rstep R')^* }" (is "_ \<subseteq> ?R'")
proof -
  {
    fix t s
    assume s: "s \<in> L" and steps: "(s,t) \<in> (rstep R)^*"
    from steps s 
    have "t \<in> ?R'"
    proof (induct)
      case base
      from base_lift[of s] base
      show ?case by force
    next
      case (step t u)
      from step(3)[OF step(4)] have "t \<in> ?R'" .
      then obtain t' s' where s': "s' \<in> lift ` L"
        and t: "t = base t'"
        and steps: "(s',t') \<in> (rstep R')^*" by auto
      from lift[OF step(2) t] obtain u' where u: "u = base u'"
        and step: "(t',u') \<in> rstep R'" by auto
      show ?case
        by (rule, rule u, rule, rule, intro conjI, rule s', insert steps step, auto)
    qed
  }
  thus ?thesis by auto
qed

fun sequence_gen :: "('g,'v)term \<Rightarrow> (nat \<Rightarrow> ('f,'v)term) \<Rightarrow> nat \<Rightarrow> ('g,'v)term" 
  where "sequence_gen s' f 0 = s'"
  | "sequence_gen s' f (Suc i) = (SOME t'. f (Suc i) = base t' \<and> (sequence_gen s' f i, t') \<in> rstep R' )"

lemma enrichment3: assumes LL: "lift ` L \<subseteq> L'"
  and SN: "\<And> t'. t' \<in> L' \<Longrightarrow> SN_on (rstep R') {t'}"
  and tL: "t \<in> L"
  shows "SN_on (rstep R) {t}"
proof
  fix f
  assume t: "f 0 \<in> {t}" and steps: "\<forall> i. (f i, f (Suc i)) \<in> rstep R"
  hence steps: "\<And> i. (f i, f (Suc i)) \<in> rstep R" by auto
  from LL tL have tL': "lift t \<in> L'" by auto
  from base_lift[of t]  tL have bl: "base (lift t) = t" by auto
  {
    fix i
    have "f (Suc i) = base (sequence_gen (lift t) f (Suc i)) \<and> (sequence_gen (lift t) f i, sequence_gen (lift t) f (Suc i)) \<in> rstep R'"
    proof (induct i)
      case 0
      thus ?case 
        by (simp add: bl singletonD[OF t], insert someI_ex[OF
          lift[OF steps[of 0, unfolded singletonD[OF t]] bl[symmetric]]], auto)
    next
      case (Suc i)
      from someI_ex[OF lift[OF steps[of "Suc i"] Suc[THEN conjunct1]]]
      show ?case by simp
    qed      
    note this[THEN conjunct2]
  } note steps = this
  have "sequence_gen (lift t) f 0 \<in> L'" by (simp add: tL')
  with steps SN[OF this] show False unfolding SN_on_def
    by blast
qed
end


fun base :: "'f \<times> nat \<Rightarrow> 'f" where "base (f,h) = f"
fun height :: "'f \<times> nat \<Rightarrow> nat" where "height (f,h) = h"
fun lift :: "nat \<Rightarrow> 'f \<Rightarrow> 'f \<times> nat" where "lift h f = (f,h)"
abbreviation base_term where "base_term \<equiv> map_term base"
abbreviation lift_term where "lift_term h \<equiv> map_term (lift h)"

lemma base_lift: "base_term (lift_term h s) = s"
  by (simp only: map_term_comp o_def, simp add: map_term_ident[unfolded id_def])

fun map_match :: "('f,'v)term \<Rightarrow> ('g,'v)term \<Rightarrow> ('g,'v)term"
where "map_match (Var x) s' = (Var x)"
  |   "map_match (Fun f ls) (Fun g ss') = (Fun g (map (\<lambda> (l,s'). map_match l s') (zip ls ss')))"


lemma vars_term_map_match: "map_term fg s' = s \<cdot> \<sigma> \<Longrightarrow> vars_term (map_match s s') = vars_term s"
proof (induct s arbitrary: s', simp)
  case (Fun f ss)
  then obtain g ss' where s': "s' = Fun g ss'" by (cases s', auto)
  from Fun(2)[unfolded s'] have ss': "map (map_term fg) ss' = map (\<lambda> t. t \<cdot> \<sigma>) ss" by (auto)
  from ss' have "length (map (map_term fg) ss') = length (map (\<lambda> t. t \<cdot> \<sigma>) ss)" by simp
  hence len: "length ss = length ss'" by simp
  {
    fix i
    assume i: "i < length ss"
    hence mem: "ss ! i \<in> set ss" by auto
    from map_nth_conv[OF ss'] i len have "map_term fg (ss' ! i) = ss ! i \<cdot> \<sigma>" by auto
    with Fun(1)[unfolded s', OF mem this] have "vars_term (ss ! i) = vars_term (map_match (ss ! i) (ss' ! i))" by auto
  } note main = this
  have id: "set ss = set (map (\<lambda> i. ss ! i) [0 ..< length ss])" 
    using map_nth[of ss] by auto
  show ?case unfolding s' 
    by (simp add: zip_nth_conv[OF len] id, insert main len, auto)
qed



(* lemma 4, existence *)
lemma map_term_linear_ex: assumes lin: "linear_term l"
  and inst: "map_term fg s' = l \<cdot> \<sigma>"
  shows "map_term fg (map_match l s') = l \<and> (\<exists> \<tau>. s' = (map_match l s') \<cdot> \<tau>)"
using assms
proof (induct l arbitrary: s')
  case (Var x)
  show ?case
    by (simp, rule exI[of _ "(\<lambda> _. s')"], simp)
next
  case (Fun f ss)
  from Fun(3) obtain f' ss' where s': "s' = Fun f' ss'" and 
    f': "fg f' = f" and ss': "map (map_term fg) ss' = map (\<lambda> t. t \<cdot> \<sigma>) ss" by (cases s', auto)
  from ss' have "length (map (map_term fg) ss') = length (map (\<lambda> t. t \<cdot> \<sigma>) ss)" by simp
  hence len: "length ss = length ss'" by simp
  have l2: "length [0 ..< length ss] = length ss" by auto
  have l3: "length [0 ..< length ss'] = length ss'" by auto
  {
    fix i
    assume i: "i < length ss"
    hence mem: "ss ! i \<in> set ss" by auto
    with Fun(2) have lin: "linear_term (ss ! i)" by auto
    from map_nth_conv[OF ss'] i len have "map_term fg (ss' ! i) = ss ! i \<cdot> \<sigma>" by auto
    note Fun(1)[OF mem lin this] this
  } note ind = this
  let ?P = "\<lambda> i \<tau>. map_term fg (map_match (ss ! i) (ss' ! i)) = ss ! i \<and> 
                  ss' ! i = map_match (ss ! i) (ss' ! i) \<cdot> \<tau>"
  from ind have "\<forall> i. \<exists> \<tau>. i < length ss \<longrightarrow> ?P i \<tau>" by blast
  from choice[OF this] obtain \<tau> where \<tau>: "\<And> i. i < length ss \<Longrightarrow> ?P i (\<tau> i)" by auto
  from Fun(2) have part_ss: "is_partition (map vars_term ss)" by  auto
  have "map vars_term ss = map vars_term (map (\<lambda> i. map_match (ss ! i) (ss' ! i)) [0 ..< length ss])"
  proof (simp add: o_def, rule nth_map_conv, simp, simp, intro allI impI)
    fix i
    assume i: "i < length ss"
    from vars_term_map_match[OF ind(2)[OF i]]
    show "vars_term (ss ! i) = vars_term (map_match (ss ! i) (ss' ! i))" ..
  qed
  from subst_merge[OF part_ss[unfolded this], of "\<lambda> i. (\<tau> i)"]
  obtain \<tau>2 where \<tau>2: "\<And> i x. i < length ss \<Longrightarrow> x \<in> vars_term (map_match (ss ! i) (ss' ! i)) \<Longrightarrow> \<tau>2 x = \<tau> i x" by auto
  show ?case unfolding s' 
  proof (intro conjI, simp add: f' o_def zip_nth_conv[OF len] len[symmetric], unfold map_nth_eq_conv[OF l2], simp add: ind)
    show "\<exists> \<tau>. Fun f' ss' = map_match (Fun f ss) (Fun f' ss') \<cdot> \<tau>"
    proof (simp add: o_def zip_nth_conv[OF len], unfold eq_commute[of ss']
      map_nth_eq_conv[OF l3], simp)
      show "\<exists> \<tau>. \<forall>i < length ss'. map_match (ss ! i) (ss' ! i) \<cdot> \<tau> = ss' ! i"
      proof (rule exI[of _ "\<tau>2"], intro allI impI)
        fix i
        assume "i < length ss'"
        with len have i: "i < length ss" by auto
        have "map_match (ss ! i) (ss' ! i) \<cdot> \<tau>2 = 
              map_match (ss ! i) (ss' ! i) \<cdot> \<tau> i" 
          by (rule term_subst_eq, insert \<tau>2[OF i], auto)
        also have "... = ss' ! i" by (rule \<tau>[OF i, THEN conjunct2, symmetric])
        finally show "map_match (ss ! i) (ss' ! i) \<cdot> \<tau>2  = ss' ! i" .
      qed
    qed
  qed
qed

(* lemma 4, uniqueness *)
lemma map_term_linear_unique: assumes lin: "linear_term l"
  and inst: "map_term fg s' = l \<cdot> \<sigma>"
  and base: "map_term fg t' = l"
  and match: "s' = t' \<cdot> \<tau>"
  shows "t' = map_match l s'"
using assms 
proof (induct l arbitrary: s' t')
  case (Var x)
  hence "t' = Var x" by (cases t', auto)
  with Var show ?case by auto
next
  case (Fun f ss)
  from Fun(3)
  obtain f' ss' where s': "s' = Fun f' ss'" and 
    f': "fg f' = f" and ss': "map (map_term fg) ss' = map (\<lambda> t. t \<cdot> \<sigma>) ss" by (cases s', auto)
  from ss' have "length (map (map_term fg) ss') = length (map (\<lambda> t. t \<cdot> \<sigma>) ss)" by simp
  hence len: "length ss = length ss'" by simp
  from Fun(4) obtain g' ts' where t': "t' = Fun g' ts'" by (cases t', auto)
  from Fun(5)[unfolded s' t'] t' have t': "t' = Fun f' ts'" and len2: "length ss' = length ts'" and ts': "ss' = map (\<lambda> t. t \<cdot> \<tau>) ts'" by auto
  have len3: "length [0 ..< length ts' ] = length ts'" by auto
  {
    fix i
    assume  "i < length ts'"
    hence i: "i < length ss" using len len2 by simp
    hence mem: "ss ! i \<in> set ss" by auto
    with Fun(2) have lin: "linear_term (ss ! i)" by auto
    from i len map_nth_conv[OF ss'] 
    have base: "map_term fg (ss' ! i) = ss ! i \<cdot> \<sigma>" by auto
    from ts'[unfolded eq_commute[of ss'] map_nth_eq_conv[OF len2[symmetric]]]
      i len have ssi: "ss' ! i = ts' ! i \<cdot> \<tau>" by auto
    from Fun(4)[unfolded t'] have id: "map (map_term fg) ts' = ss" by simp
    from len len2 have "length ts' = length ss" by simp
    from id[unfolded map_nth_eq_conv[OF this]] i have base2: "map_term fg (ts' ! i) = ss ! i" by simp
    have "ts' ! i = map_match (ss ! i) (ss' ! i)" 
      by (rule Fun(1)[OF mem lin base base2 ssi])
  } note ind = this 
  show ?case unfolding t' s' 
    by (simp add: zip_nth_conv[OF len] o_def eq_commute[of ts'] len2
      map_nth_eq_conv[OF len3], insert ind, auto)
qed


fun sym_collect :: "(('f,'v)term \<Rightarrow> bool) \<Rightarrow> ('f,'v)term \<Rightarrow> 'f list"
where "sym_collect p (Var x) = []"
  |   "sym_collect p (Fun f ts) = (if p (Fun f ts) then [f] else []) @ 
  concat (map (sym_collect p) ts)"

definition cover :: "(('f,'v)rule \<Rightarrow> ('f,'v)term \<Rightarrow> bool) \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f \<times> nat,'v)trs"
  where "cover ff R \<equiv> { (l', lift_term h r) | l r l' h. 
       (l,r) \<in> R 
     \<and> base_term l' = l 
     \<and> h = Suc (min_list (map height (sym_collect (\<lambda> t'. ff (l,r) (base_term t')) l')))}"

lemma cover_left_linear: assumes "left_linear_trs R"
  shows "left_linear_trs (cover ff R)"
  using assms
  unfolding left_linear_trs_def cover_def
  by (clarify, insert linear_map_term[of base], auto)

lemma cover_var_condition: assumes "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and "(l,r) \<in> cover ff R"
  shows "vars_term r \<subseteq> vars_term l"
  using assms vars_map_term2[of base]
  unfolding cover_def
  by force

lemma cover6a: "(l',r') \<in> cover ff {(l,r)} \<Longrightarrow> (base_term l',base_term r') = (l,r)"
  unfolding cover_def
  by (auto simp: base_lift)
  
lemma cover6b: assumes step: "(s',t') \<in> rstep (cover ff R)"
  shows "(base_term s',base_term t') \<in> rstep R"
  using assms
proof (induct)
  case (IH C \<sigma> l' r')
  then obtain l r where lr: "(l,r) \<in> R" and l'r': "(l',r') \<in> cover ff {(l,r)}"
    unfolding cover_def by auto
  from lr cover6a[OF l'r'] have lr: "(base_term l',base_term r') \<in> R" by auto
  show ?case
    by (simp, rule, rule lr, auto)
qed

(* proposition 7 extended (the original formulation does not suffice
   to get lemma 13) *)
lemma enrichment_cover_extended: assumes lin: "linear_term l"
   and wf: "vars_term r \<subseteq> vars_term l"
   and s': "base_term s' = C\<langle>l \<cdot> \<sigma>\<rangle>"
   shows "\<exists> l' r' \<tau> D. (l',r') \<in> cover ff {(l,r)} \<and> s' = D\<langle>l' \<cdot> \<tau>\<rangle> \<and> C = map_ctxt base D \<and> l = base_term l' \<and> r = base_term r' \<and> (\<forall> x \<in> vars_term r. (map_subst base \<tau>) x = \<sigma> x)"
proof -
  from map_term_ctxt_decomp[OF s'] obtain D u' where
      D: "C = map_ctxt base D" and base: "base_term u' = l \<cdot> \<sigma>" 
      and s': "s' = D\<langle>u'\<rangle>" by auto
  from map_term_linear_ex[OF lin base] obtain l' \<tau> where l: "l = base_term l'"
      and u': "u' = l' \<cdot> \<tau>" by force
  from l obtain r' where coverlr: "(l',r') \<in> cover ff {(l,r)}" 
    unfolding cover_def by auto
  from cover6a[OF coverlr] have r: "r = base_term r'" by auto      
  show ?thesis unfolding s' u' 
  proof (rule exI[of _ l'], rule exI[of _ r'], rule exI[of _ \<tau>], rule exI[of _ D], simp add: coverlr D, simp add: l r, intro ballI)
    fix x
    assume "x \<in> vars_term r'"
    thus "base_term (\<tau> x) = \<sigma> x"
      using term_subst_eq_rev[OF base[unfolded l s' u', simplified]] wf
        unfolding l r vars_map_term2 by auto
  qed
qed

lemma enrichment_cover: assumes ll: "left_linear_trs R"
  and wf: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  shows "enrichment R (cover ff R) (lift_term n) base_term"
proof
  fix t
  show "base_term (lift_term n t) = t" by (rule base_lift)
next
  fix s t s'
  assume step: "(s,t) \<in> rstep R" and s: "s = base_term s'"
  thus "\<exists> t'. t = base_term t' \<and> (s',t') \<in> rstep (cover ff R)"
  proof (induct arbitrary: s')
    case (IH C \<sigma> l r s')
    from IH(1) ll have lin:"linear_term l" unfolding left_linear_trs_def by auto
    from IH(2) have s': "base_term s' = C\<langle>l \<cdot> \<sigma>\<rangle>" by simp    
    from enrichment_cover_extended[OF lin wf[OF IH(1)] s']
    obtain l' r' \<tau> D where rule: "(l',r') \<in> cover ff {(l,r)}"
      and s': "s' = D\<langle>l' \<cdot> \<tau>\<rangle>" and D: "C = map_ctxt base D"
      and l: "l = base_term l'" and r: "r = base_term r'"
      and \<tau>: "\<And> x. x \<in> vars_term r \<Longrightarrow> (map_subst base \<tau>) x = \<sigma> x" by blast+
    show ?case
      by (rule exI[of _ "D\<langle>r' \<cdot> \<tau>\<rangle>"], unfold l r D s', intro conjI, simp, rule term_subst_eq, rule \<tau>[symmetric,simplified], insert rule IH(1), unfold cover_def, auto)
  qed
qed


(* definition 12 and lemma 13 of Matchbounds paper are available but not required *)

fun roof :: "('f,'v)rule \<Rightarrow> ('f,'v) term \<Rightarrow> bool"
  where [code del]: "roof (l,r) = (\<lambda> t. vars_term r \<subseteq> vars_term t)"


lemma height_lift: assumes "(f,n) \<in> funas_term (lift_term h t)"
  shows "height f = h"
using assms
by (induct t, auto)

lemma roof_locally_SN: assumes wf_trs: "wf_trs R" shows "locally_terminating (cover roof (R :: ('f,'v)trs))"
  unfolding locally_terminating_def
proof (intro allI impI)
  fix F :: "('f \<times> nat)sig"
  assume "finite F"
  from finite_list[OF this] obtain Fs where F: "F = set Fs" by auto
  let ?hs = "map (\<lambda>(f,n). height f) Fs"
  obtain m where m: "m = Suc (max_list ?hs)" by auto
  have small: "\<And> f n. (f,n) \<in> F \<Longrightarrow> height f < m" unfolding F m using max_list[of _ ?hs] by force
  let ?S = "sig_step F (rstep (cover roof R))"
  let ?R = "sig_step F (cover roof R)"
  have subset: "?S \<subseteq> rstep ?R" 
  proof 
    fix s t
    assume "(s,t) \<in> ?S"
    from sig_stepE[OF this] have sF: "funas_term s \<subseteq> F" and tF: "funas_term t \<subseteq> F" and step: "(s,t) \<in> rstep (cover roof R)" by auto
    from step obtain C l r \<sigma> where lr: "(l,r) \<in> cover roof R" 
      and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
    from sF tF have F: "funas_term l \<subseteq> F" "funas_term r \<subseteq> F" unfolding s t 
      by (auto simp: funas_term_subst)
    from sig_stepI[OF lr this]
    show "(s,t) \<in> rstep ?R" unfolding s t by auto
  qed
  let ?prc = "\<lambda> (f :: 'f \<times> nat) n g k. (height f < height g \<and> height g \<le> m, height f < height g \<and> height g \<le> m \<or> f = g \<and> n = k)"
  let ?prl = "\<lambda> _. False"
  interpret rpo_pr ?prc ?prl
  proof(unfold_locales, unfold fst_conv)
    show "SN {((f,n), (g,k)) | f n g k. height f < height g \<and> height g \<le> m}"
      by (rule SN_subset[OF SN_inv_image[of _ "\<lambda>(f,n). height f", unfolded inv_image_def, OF SN_nat_bounded[of m]]], auto)
  qed auto
  show "SN ?S"
  proof (rule SN_subset[OF _ subset], rule rpo_manna_ness, unfold sig_step_def, clarify)
    fix l r
    assume lr: "(l,r) \<in> cover roof R"
    and wfl: "funas_term l \<subseteq> F" and wfr: "funas_term r \<subseteq> F"
    from lr[unfolded cover_def]
    obtain ll rr h where base: "base_term l = ll" and llrr: "(ll,rr) \<in> R" and
      r: "r = lift_term h rr" and
      h: "h = Suc (min_list (map height (sym_collect (\<lambda> t'. roof (ll,rr) (base_term t')) l)))" by auto
    from wf_trs[unfolded wf_trs_def] llrr obtain ff lls where ll: "ll = Fun ff lls" and vars: "vars_term rr \<subseteq> vars_term ll" 
      by (cases ll, force+)
    from vars r base vars_map_term2[of "lift h" rr] vars_map_term2[of base l] 
    have vars: "vars_term r \<subseteq> vars_term l" and vars2: "vars_term rr = vars_term r" by auto
    obtain vr where vr: "vars_term r = vr" by auto
    from base ll have "is_Fun l" by (cases l, auto)
    with vars have "\<exists> f' ls'. l \<unrhd> Fun f' ls' \<and> Suc (height f') = h \<and> vars_term r \<subseteq> vars_term (Fun f' ls')" unfolding h roof.simps vars2[symmetric] vars_map_term2 unfolding vars2 vr nat.inject
    proof (induct l)
      case (Var x) thus ?case by simp
    next
      case (Fun f ls) 
      let ?fs = "\<lambda> s. sym_collect (\<lambda> t'. vr \<subseteq> vars_term t') s"
      let ?hs = "\<lambda> s. map height (?fs s)"
      let ?h = "\<lambda> s. min_list (?hs s)"
      show ?case 
      proof (cases "height f = ?h (Fun f ls)")
        case True
        show ?thesis unfolding True[symmetric]
          by (intro exI conjI, rule supteq.refl, rule, rule Fun(2))
      next
        case False        
        let ?fls = "concat (map ?fs ls)"
        from Fun(2) have id: "?fs (Fun f ls) = f # ?fls" by auto
        hence id: "?hs (Fun f ls) = height f # map height ?fls" by auto
        hence hsnNil: "?hs (Fun f ls) \<noteq> []" by simp
        from min_list_ex[OF hsnNil] obtain h where mem: "h \<in> set (?hs (Fun f ls))"
          and min: "h = min_list (?hs (Fun f ls))" by auto
        from False mem min have "h \<in> set (map height ?fls)" unfolding id by auto
        then obtain l where l: "l \<in> set ls" and h: "h \<in> set (map height (?fs l))"
          by auto
        have subset: "set (map height (?fs l)) \<subseteq> set (?hs (Fun f ls))" using l by auto
        from h have vr: "vr \<subseteq> vars_term l" by (induct l, auto)
        from h have var: "is_Fun l" by (cases l, auto)
        from Fun(1)[OF l vr var] obtain f' ls' where supteq: "l \<unrhd> Fun f' ls'"
          and hf': "height f' = min_list (?hs l)" and vr: "vr \<subseteq> vars_term (Fun f' ls')"
          by auto
        from l supteq have supteq: "Fun f ls \<unrhd> Fun f' ls'" by auto
        show ?thesis 
          by (intro exI conjI, rule supteq, unfold hf' min_list_subset[OF subset h[unfolded min]], insert vr, auto)
      qed
    qed      
    then obtain f' ls' where sub: "l \<unrhd> Fun f' ls'" and h: "h = Suc (height f')" and vars: "vars_term r \<subseteq> vars_term (Fun f' ls')" by auto
    from wfl sub have "(f',length ls') \<in> F" by auto
    from small[OF this] have f'm: "height f' < m" .
    {
      fix g k
      assume "(g,k) \<in> funas_term r"
      from height_lift[OF this[unfolded r]]
      have "height g = h" by simp        
      with h f'm have "height f' < height g \<and> height g \<le> m" by simp
    } note main = this
    have "rpo_pr 0 (Fun f' ls') r = (True,True)" 
      by (rule rpo_large_sym[OF _ vars], insert main, auto)    
    from supteq_imp_rpo_nstri[OF sub] this
    show "fst (rpo_pr 0 l r)" using rpo_compat by force
  qed
qed


definition e_bounded :: "(('f,'v)rule \<Rightarrow> ('f,'v)term \<Rightarrow> bool) \<Rightarrow> nat \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)term set \<Rightarrow> bool"
  where "e_bounded e c R L \<equiv> \<forall> s t f n. s \<in> lift_term 0 ` L \<and> (s,t) \<in> (rstep (cover e R))^* \<and> (f,n) \<in> funas_term t \<longrightarrow> height f \<le> c"



(* lemma 22a *)
lemma e_bounded_compact: assumes wf_trs: "wf_trs R"
  and bounded: "e_bounded e c R L"
  and L: "\<Union> funas_term ` L \<subseteq> F"
  and R: "finite R"
  and F: "finite F"
  shows "compact (lift_term 0 ` L) (cover e R)"
proof -
  from finite_list[OF R] obtain Rl where R: "R = set Rl" by auto
  have F': "finite (funas_trs R)" unfolding R funas_trs_list_sound[of Rl, symmetric]
    by (rule finite_set)  
  let ?F' = "funas_trs R" 
  let ?G = "F \<union> ?F'"
  obtain G where G: "G = ?G" by auto
  from F F' have finG: "finite G" unfolding G by auto
  let ?F = "{((f,h),n). (f,n) \<in> G \<and> h \<le> c}"
  from finG have "finite (G \<times> {h. h \<le> c})" (is "finite ?H") by auto
  hence fin: "finite ((\<lambda> ((f,n),h). ((f,h),n)) ` ?H)" by auto
  have F: "finite ?F"
    by (rule HOL.subst[where P = finite,  OF _ fin], force)  
  {
    fix s t 
    assume steps: "(lift_term 0 s, t) \<in> (rstep (cover e R))^*" and sL: "s \<in> L"
    with bounded[unfolded e_bounded_def] have bound: "\<And> f n. (f,n) \<in> funas_term t \<Longrightarrow> height f \<le> c" by blast
    from rsteps_imp_map_rsteps[OF steps, of base, unfolded base_lift]
    have steps: "(s,base_term t) \<in> (rstep (map_trs base (cover e R)))^*" (is "_ \<in> (rstep ?R)^*") .
    { 
      fix l r 
      assume "(l,r) \<in> ?R"
      then obtain l' r' where l'r': "(l',r') \<in> cover e R" and l: "l = base_term l'"
        and r: "r = base_term r'" unfolding map_trs.simps by force
      then obtain l'' r'' n where rule: "(l'',r'') \<in> R" and l': "base_term l' = l''" and r': "r' = lift_term n r''" unfolding cover_def by auto
      have "(l,r) \<in> R" unfolding l r l' r' base_lift by (rule rule)
    }          
    hence subset: "?R \<subseteq> R" by auto
    from steps rstep_mono[OF subset] have steps:"(s, base_term t) \<in> (rstep R)^*"
      by (induct, auto)
    from sL L have "funas_term s \<subseteq> F" by auto
    hence sG: "funas_term s \<subseteq> G" using F G unfolding G by blast    
    have FG: "funas_trs R \<subseteq> G" unfolding G by auto
    from rsteps_preserve_funas_terms[OF FG sG steps wf_trs]     
    have base:"funas_term (base_term t) \<subseteq> G" .
    from base bound
    have "funas_term t \<subseteq> ?F" 
    proof (induct t)
      case (Var x) thus ?case by auto
    next
      case (Fun f ts)
      from Fun(2)[unfolded map_term.simps] have bf: "(base f,length ts) \<in> G"         
        by simp
      from Fun(3)[of f "length ts"] have hf: "height f \<le> c" by auto
      from bf hf have f: "(f,length ts) \<in> ?F" by auto
      {
        fix t
        assume t: "t \<in> set ts"
        from t Fun(3) have h: "\<And> f n. (f,n) \<in> funas_term t \<Longrightarrow> height f \<le> c" by auto
        from t Fun(2) have b: "funas_term (base_term t) \<subseteq> G" by auto
        from Fun(1)[OF t b h]
        have "funas_term t \<subseteq> ?F" by auto
      } note ind = this
      from f ind show ?case by auto
    qed
  } note main = this
  show ?thesis
    by (unfold compact_def, intro exI conjI, rule F, insert main, auto)
qed




(* theorem 23 *)
lemma e_bounded: assumes wf_trs: "wf_trs R"
  and SN: "locally_terminating (cover e R)"
  and R: "finite R"
  and F: "finite F"
  and L: "\<Union> funas_term ` L \<subseteq> F"
  and ll: "left_linear_trs R"
  and bounded: "e_bounded e c R L"
  and t: "t \<in> L"
  shows "SN_on (rstep R) {t}"
proof -
  from e_bounded_compact[OF wf_trs bounded L R F]
  have compact: "compact (lift_term 0 ` L) (cover e R)" .
  interpret enrichment R "cover e R" "lift_term 0" base_term
    by (rule enrichment_cover[OF ll], insert wf_trs, unfold wf_trs_def, auto)
  show ?thesis
  proof (rule enrichment3[OF subset_refl _ t])
    fix t'
    assume "t' \<in> lift_term 0 ` L"
    from locally_terminating_compact[OF SN compact] this
    show "SN_on (rstep (cover e R)) {t'}" by auto
  qed
qed


definition match :: "('f,'v)rule \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
  where "match \<equiv> \<lambda> _ _. True"


lemma sym_collect_match: "sym_collect (\<lambda> t'. match lr (f t')) t = funs_term_list' t" unfolding match_def
proof (induct t)
  case (Var x) show ?case by (simp add: funs_term_list'.simps)
next
  case (Fun f ts)
  show ?case 
    by (simp add: funs_term_list'.simps, insert Fun, induct ts, auto simp: funs_term_list'.simps)
qed

    

lemma match_lSN: assumes wf: "wf_trs (R :: ('f,'v)trs)"
  and non_duplication: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term_ms r <= vars_term_ms l"
  shows "locally_terminating (cover match R)"
unfolding locally_terminating_def
proof (intro allI impI)
  fix F :: "('f \<times> nat)sig"
  let ?gt = "\<lambda> s t. (s,t) \<in> rstep (cover match R) \<and> funas_term s \<subseteq> F \<and> funas_term t \<subseteq> F"  
  assume "finite F"
  from finite_list[OF this] obtain Fs where F: "F = set Fs" by auto
  let ?hs = "map (\<lambda>(f,n). height f) Fs"
  obtain m where m: "m = Suc (max_list ?hs)" by auto
  let ?m = "\<lambda> s. funs_term_ms (map_term (\<lambda> f. m - height f) s)"
  let ?lt = "{(x,y :: nat). x < y}"
  let ?le = "\<lambda> s t. (?m s, ?m t) \<in> mult ?lt"
  let ?R = "{(s,t). ?le s t}"
  have small: "\<And> f n. (f,n) \<in> F \<Longrightarrow> height f < m" unfolding F m using max_list[of _ ?hs] by force
  have trans: "trans ?lt" unfolding trans_def by auto
  have main: "wf {(s,t). ?gt t s}" 
  proof (rule wf_subset)
    show "wf ?R"
      using wf_inv_image[OF wf_mult[OF wf_less], of ?m]
      unfolding inv_image_def .
  next    
    {
      fix s t
      assume "?gt t s"
      hence ts: "(t,s) \<in> rstep (cover match R)"
        and t: "funas_term t \<subseteq> F"
        and s: "funas_term s \<subseteq> F" by auto      
      hence "?le s t" 
      proof (induct)
        case (IH C \<sigma> l r)
        let ?C = "funs_ctxt_ms (map_ctxt (\<lambda> f. m - height f) C)"
        let ?\<sigma> = "\<lambda> t. Multiset2.concat (Multiset2.map (\<lambda> x. funs_term_ms ((map_subst (\<lambda>f. m - height f) \<sigma>) x)) (vars_term_ms t))"
        have id: "\<And> t. ?m (C\<langle>t \<cdot> \<sigma>\<rangle>) = ?C + ?m t + ?\<sigma> t"
          by (simp  add: funs_ms_ctxt_apply funs_ms_subst_apply multiset_eq_iff)
        from cover_var_condition[OF _ IH(1)] wf[unfolded wf_trs_def]
        have vars: "vars_term r \<subseteq> vars_term l" by auto
        from IH(1)[unfolded cover_def] obtain ll rr h where base: "base_term l = ll" and llrr: "(ll,rr) \<in> R" and
          r: "r = lift_term h rr" and
          h: "h = Suc (min_list (map height (sym_collect (\<lambda> t'. match (ll,rr) (base_term t')) l)))" by auto
        let ?h = "min_list (map height (funs_term_list' l))"
        from h[unfolded sym_collect_match]
        have hS: "h = Suc (?h)" by simp
        from wf[unfolded wf_trs_def] llrr obtain ff lls where ll: "ll = Fun ff lls" by (cases ll, force+)
        {
          let ?mmm = "\<lambda>ls. multiset_of (map (\<lambda>f. m - height f) ls)"
          let ?fl = "funs_term_list' l"
          let ?mm = "min_list (map height ?fl)"
          from base ll have ne: "map height ?fl \<noteq> []" by (cases l, auto simp: funs_term_list'.simps)
          from min_list_ex[OF this] obtain f where
            f: "f \<in> set ?fl" and hf: "?mm = height f" by auto
          from f[unfolded funs_term_list'_sound funs_term_funas_term] 
          obtain n where fn: "(f,n) \<in> funas_term l" by auto
          from IH(2) have l: "funas_term l \<subseteq> set Fs" unfolding F by (auto simp: funas_term_subst)
          from fn l have "(f,n) \<in> set Fs" by auto
          hence hfF: "height f \<in> set (map (\<lambda> (f,n). height f) Fs)" by auto
          have m: "?h < m" unfolding m hf using max_list[OF hfF] by auto
          from f[unfolded set_conv_nth] obtain i where i: "i < length ?fl" 
            and f: "?fl ! i = f" by auto
          from id_take_nth_drop[OF i, unfolded f] obtain bef aft 
            where fl: "?fl = bef @ f # aft" by auto        
          have "\<exists> L. ?m l = L + {# m - ?h #}" unfolding multiset_of_funs_term_list'[symmetric] 
            unfolding hf funs_term_list'_map_term
            unfolding fl
            by (rule exI[of _ "?mmm bef + ?mmm aft"], simp add: multiset_eq_iff)
          note this and m
        } note Lm = this
        then obtain L where L: "?m l = L + {# m - ?h #}" by auto
        {
          fix h'
          assume h': "h' \<in># ?m r" 
          from this[unfolded multiset_of_funs_term_list'[symmetric]]
          have "h' \<in> (\<lambda>f. m - height f) ` (funs_term r)" 
            unfolding funs_term_list'_map_term in_multiset_in_set using funs_term_list'_sound by auto
          from this[unfolded funs_term_funas_term]  
          obtain f n where f: "(f,n) \<in> funas_term r" and h': "h' = m - height f"
            by auto
          from height_lift[OF f[unfolded r]] h' have h': "h' = m - h" by auto
          have "h' < m - ?h" unfolding h' hS using Lm(2) by auto
        }
        with L obtain h where l: "?m l = L + {#h#}" and h: "\<forall>h'. h' \<in># ?m r \<longrightarrow> h' < h" by auto
        from non_duplication[OF llrr] have "vars_term_ms (base_term r) <= vars_term_ms (base_term l)" unfolding base r by auto 
        hence "vars_term_ms r <= vars_term_ms l" by simp
        then obtain M where Mid: "vars_term_ms l = vars_term_ms r + M" 
          unfolding mset_le_exists_conv ..
        obtain \<sigma>diff where \<sigma>l: "?\<sigma> l = ?\<sigma> r + \<sigma>diff" unfolding Mid
          by auto
        have l: "?m (C\<langle>l \<cdot> \<sigma>\<rangle>) = (?C + ?\<sigma> r) + (L + {#h#} + \<sigma>diff)"
          unfolding id l \<sigma>l multiset_eq_iff by auto
        have r: "?m (C\<langle>r \<cdot> \<sigma>\<rangle>) = (?C + ?\<sigma> r) + ?m r"
          unfolding id multiset_eq_iff by auto
        show "?le (C\<langle>r \<cdot> \<sigma>\<rangle>) (C\<langle>l \<cdot> \<sigma>\<rangle>)"        
          unfolding l r          
          by (rule one_step_implies_mult[OF trans], insert h, auto)
      qed
    }
    thus "{(s,t). ?gt t s} \<subseteq> ?R" by auto
  qed
  term ?gt
  have id: "{(s,t). ?gt t s} = (sig_step F (rstep (cover match R)))^-1" 
    unfolding sig_step_def by auto
  show "SN (sig_step F (rstep (cover match R)))"
    by (rule wf_imp_SN[OF main[unfolded id]])
qed

hide_const (open) match
hide_const (open) roof

end
