(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Multiset2
imports
  "~~/src/HOL/Library/Multiset"
  Relations
  List_Util
begin


lemma in_multiset_of: assumes "x \<in># multiset_of xs"
  shows "\<exists> bef aft. xs = bef @ x # aft"
proof -
  from assms[unfolded in_multiset_in_set set_conv_nth]
  obtain i where i: "i < length xs" and x: "xs ! i = x" by auto
  from id_take_nth_drop[OF i, unfolded x]
  show ?thesis by auto
qed

lemma in_multiset_of_idx :
  assumes "a \<in># multiset_of as"
  shows "\<exists>i. i < length as \<and> a = as ! i"
proof-
 from assms and in_multiset_of[of as a] obtain bef aft where
   as_part: "as = bef @ a # aft" by auto
 with nth_append_length have a_eq: "a = as ! (length bef)" by auto
 from as_part and length_append have i_bd: "(length bef) < length as" by auto
 from a_eq and i_bd show ?thesis by blast
qed

lemma in_set_idx:
  assumes "a \<in> set as"
  shows "\<exists>i. i < length as \<and> a = as ! i"
using assms and in_multiset_in_set[of as a] and in_multiset_of_idx[of as a]
by simp

lemma mem_multiset_diff_compat:
  assumes "x \<in># A"
  and "x \<noteq> y"
  shows "x \<in># (A - {#y#})"
using assms by (induct A) auto

lemma in_multiset_of_eq: assumes "M + {#x#} = multiset_of ys"
  shows "\<exists> bef aft. ys = bef @ x # aft \<and> M = multiset_of (bef @ aft)"
proof -
  from assms[symmetric] have "x \<in># multiset_of ys" by auto
  from in_multiset_of[OF this] obtain bef aft where ys: "ys = bef @ x # aft"
    by auto
  have "M + {#x#} = multiset_of (bef @ aft) + {#x#}"
    unfolding assms ys multiset_eq_iff by auto
  hence "M = multiset_of (bef @ aft)" by auto
  from ys this show ?thesis by auto
qed

lemma multiset_of_map: assumes "multiset_of xs = multiset_of ys"
  shows "multiset_of (map f xs) = multiset_of (map f ys)"
  using assms
proof (induct xs arbitrary: ys)
  case Nil thus ?case by auto
next
  case (Cons x xs)
  let ?m = "multiset_of"
  from in_multiset_of_eq[OF Cons(2)[simplified]]
  obtain bef aft where ys: "ys = bef @ x # aft" and xs: "?m xs = ?m (bef @ aft)"
    by auto
  from Cons(1)[OF xs] show ?case unfolding ys by (auto simp: multiset_eq_iff)
qed

lemma multiset_of_concat: assumes "multiset_of xs = multiset_of ys"
  shows "multiset_of (concat xs) = multiset_of (concat ys)"
  using assms
proof (induct xs arbitrary: ys)
  case Nil thus ?case by auto
next
  case (Cons x xs)
  let ?m = "multiset_of"
  from in_multiset_of_eq[OF Cons(2)[simplified]]
  obtain bef aft where ys: "ys = bef @ x # aft" and xs: "?m xs = ?m (bef @ aft)"
    by auto
  from Cons(1)[OF xs] show ?case unfolding ys by (auto simp: multiset_eq_iff)
qed

definition list_of :: "'a multiset \<Rightarrow> 'a list"
  where "list_of M \<equiv> SOME xs. multiset_of xs = M"

lemma multiset_of_list_of[simp]: "multiset_of (list_of M) = M"
  by (unfold list_of_def, rule someI_ex, 
    insert surj_multiset_of[unfolded surj_def, THEN spec[of _ M]], 
    auto)

lemma list_of_singleton[simp]: "list_of {#x#} = [x]"
  unfolding list_of_def
proof(rule, simp)
  fix xs
  assume id: "multiset_of xs = {#x#}"
  then obtain y ys where xs: "xs = y # ys" by (cases xs, auto)
  show "xs = [x]" 
  proof (cases ys)
    case Nil
    with id xs show ?thesis by auto
  next
    case (Cons z zs)
    from id[unfolded xs Cons]
    have "multiset_of zs + {#z#} + {#y#} = {#x#}" by auto
    from this[unfolded union_is_single] show ?thesis by auto
  qed
qed

lemma list_of_empty[simp]: "list_of {#} = []"
  unfolding list_of_def
proof(rule, simp)
  fix xs
  assume id: "multiset_of xs = {#}"
  thus "xs = []" by (cases xs, auto)
qed
    
definition map :: "('a \<Rightarrow> 'b) \<Rightarrow> 'a multiset \<Rightarrow> 'b multiset"
  where "map f M \<equiv> multiset_of (List.map f (list_of M))"

lemma map_union[simp]: "map f (M + N) = map f M + map f N"
proof -
  let ?xs = list_of
  have "multiset_of (List.map f (?xs (M + N)))
    = multiset_of (List.map f (?xs M @ ?xs N))"
    by (rule multiset_of_map,  auto)
  also have "... = multiset_of (List.map f (?xs M)) +
    multiset_of (List.map f (?xs N))"
    by auto
  finally show ?thesis 
    unfolding map_def .
qed

lemma map_empty[simp]: "map f {#} = {#}" 
  unfolding map_def
  by simp

lemma map_singleton[simp]: "map f {#x#} = {#f x#}" 
  unfolding map_def
  by simp

lemma map_list_of_multiset_of: "multiset_of (List.map f (list_of (multiset_of xs))) = multiset_of (List.map f xs)"
  by (rule multiset_of_map, simp)


lemma map_multiset_of[simp]: "map f (multiset_of xs) = multiset_of (List.map f xs)" unfolding map_def map_list_of_multiset_of ..

lemma in_map_multiset_of[intro]:
  assumes "a \<in># A"
  shows "f a \<in># map f A"
proof-
from assms and multiset_of_list_of
 have "a \<in># multiset_of (list_of A)" by simp
from in_multiset_of_idx[OF this] obtain i where
  i_bd: "i < length (list_of A)" and
  eq_a_Ai: "a = list_of A ! i" by auto
hence i_bd': "i < length (List.map f (list_of A))" by simp
from nth_mem_multiset_of[OF i_bd'] and nth_map[OF i_bd] and map_def and eq_a_Ai
 show ?thesis by metis
qed

definition concat :: "'a multiset multiset \<Rightarrow> 'a multiset"
  where "concat M \<equiv> multiset_of (List.concat (List.map list_of (list_of M)))"

lemma concat_union[simp]: "concat (M + N) = concat M + concat N"
proof -
  let ?xs = list_of
  have "multiset_of (List.concat (List.map ?xs (?xs (M + N))))
       = multiset_of (List.concat (List.map ?xs (?xs M @ ?xs N)))"
    by (rule multiset_of_concat, rule multiset_of_map, simp)
  also have "... = multiset_of (List.concat (List.map ?xs (?xs M))) 
    + multiset_of (List.concat (List.map ?xs (?xs N)))"
    by auto
  finally show ?thesis unfolding concat_def .
qed

lemma concat_empty[simp]: "concat {#} = {#}" 
  unfolding concat_def
  by simp

lemma concat_singleton[simp]: "concat {#M#} = M" 
  unfolding concat_def
  by simp

lemma map_map: "map f (map g M) = map (f o g) M"
  unfolding map_def map_list_of_multiset_of map_map 
  by simp

lemma map_concat: "map f (concat Ms) = concat (map (map f) Ms)"
proof -
  obtain xs where xs: "list_of Ms = xs" by auto
  show ?thesis
    unfolding concat_def map_def[of f]
    unfolding map_list_of_multiset_of
    unfolding map_concat
    unfolding List.map_map o_def
  proof (rule multiset_of_concat, unfold map_def[symmetric], unfold map_map o_def, unfold map_def[of _ Ms] xs)
    oops (* unsolvable, problem: multiset_of_concat is too restrictive *)

lemma concat_mem_subset: assumes "M \<in># ms" shows "M \<le> concat ms"
  using assms
proof (induct ms)
  case empty thus ?case by simp
next
  case (add ms N)
  show ?case
  proof (cases "M = N")
    case True
    thus ?thesis by auto
  next
    case False
    with add(2) have "M \<in># ms" by auto
    from add(1)[OF this] have "M \<le> concat ms" .
    also have "... \<le> concat ms + concat {#N#}"
      by auto
    finally show ?thesis by simp
  qed
qed

lemma multiset_map_mono: assumes AB: "A \<le> B" shows "map f A \<le> map f B"
proof -
  from AB[unfolded mset_le_exists_conv] obtain C where B: "B = A + C" ..
  thus ?thesis by auto
qed

hide_fact(open) map_map
hide_const(open) map
hide_const(open) concat

lemma size_multiset_of : "size (multiset_of as) = length as"
by (induct as) auto

lemma diff_union_swap_single :
  assumes "a \<in># A"
  shows "A - {#a#} + {#b#} = A + {#b#} - {#a#}"
using assms
proof (cases "a = b")
 case True
  from this and insert_DiffM2 and assms show ?thesis by auto
 next
 case False
  from this and diff_union_swap show ?thesis by force
qed
  
lemma diff_union_swap_multiset :
  assumes "a \<in># A"
  shows "A - {#a#} + B = A + B - {#a#}"
proof (induct B)
 case empty
  thus ?case by auto
 next
 case (add B b)
  with union_assoc[of "A - {#a#}", symmetric]
   have Heq: "A - {#a#} + (B + {#b#}) = A + B - {#a#} + {#b#}" by simp
  from assms have "a \<in># (A + B)" by simp
  from Heq and diff_union_swap_single[symmetric, OF this] and union_assoc[of A B]
   show ?case by simp
qed

lemma list_empty_inv :
  assumes "list_of A = []"
  shows "A = {#}"
proof-
  from assms and multiset_of_list_of[of A] show ?thesis by auto
qed

text{* In a multiset, there exists a local maxima. *}

lemma local_maxima_multiset_aux:
  defines Hrew[simp]: "m \<equiv> multiset_of" 
  assumes "locally_trans s (m xs) (m xs) (m xs)" (is "?lt xs")
  and "locally_irrefl s (m xs)"    (is "?li xs")
  shows "xs \<noteq> [] \<longrightarrow> (\<exists>a. a \<in># (m xs) \<and> (\<forall>b. b \<in># (m xs) \<longrightarrow> (b,a) \<notin> s))"
using assms
proof (induct xs)
case Nil show ?case by auto
next
case (Cons x xs)
  from `?lt (x # xs)` and lt_trans_l have lt_xs: "?lt xs" by simp
  from `?li (x # xs)` and li_trans_l have li_xs: "?li xs" by simp
  from li_xs and lt_xs and Cons(1)
   have IH: "xs \<noteq> [] \<longrightarrow> (\<exists>a. a \<in># multiset_of xs \<and> (\<forall>b. b \<in># multiset_of xs \<longrightarrow> (b,a) \<notin> s))"
   by simp
 show ?case
  proof (cases xs)
  case Nil
   with `?li (x # xs)`[unfolded locally_irrefl_def, rule_format] show ?thesis
    by simp
  next
  case (Cons y ys)
   from this and IH obtain a1 where
    a1_in_xs: "a1 \<in># multiset_of xs"
    and a1_is_lm: "\<forall>b. b \<in># multiset_of xs \<longrightarrow> (b,a1) \<notin> s" by force
   show ?thesis 
    proof (cases "(x,a1) \<in> s")
    case True note sxa1 = this
     have GL: "x \<in># multiset_of (x # xs)" by auto
     from `?li (x # xs)`[unfolded locally_irrefl_def]
      have GR1: "(x,x) \<notin> s" by simp
     {
       fix b assume b_in_xs: "b \<in># multiset_of xs"
       have "(b,x) \<notin> s"
       proof (cases "(b,x) \<in> s")
        case True
         from b_in_xs have b_in: "b \<in># multiset_of (x # xs)" by force
         from a1_in_xs have a1_in: "a1 \<in># multiset_of (x # xs)" by force
         from b_in and a1_in and GL and True and sxa1 and 
          `?li (x # xs)`[unfolded locally_irrefl_def, rule_format] and
          `?lt (x # xs)`[unfolded locally_trans_def, rule_format]
          have "(b,a1) \<in> s" unfolding Hrew by metis
         with a1_is_lm and b_in_xs show ?thesis by auto
        next
        case False thus ?thesis by auto
        qed
     }
     hence GR2: "\<forall>b. b \<in># multiset_of xs \<longrightarrow> (b,x) \<notin> s" by simp
     from this and GL and GR1 show ?thesis by auto
    next
    case False
     from this and a1_is_lm
     have GR: "\<forall>b. b \<in># multiset_of (x # xs) \<longrightarrow> (b,a1) \<notin> s" by simp
     from a1_in_xs have GL: "a1 \<in># multiset_of (x # xs)" by simp
     from GL and GR show ?thesis unfolding Hrew by metis
    qed
  qed
qed

lemma local_maxima_multiset :
  assumes "locally_trans s A A A" (is "?lt A")
    and "locally_irrefl s A" (is "?li A")
    and A_ne: "A \<noteq> {#}"
  shows "\<exists>a. a \<in># A \<and> (\<forall>b. b \<in># A \<longrightarrow> (b,a) \<notin> s)"
proof-
 from A_ne have lA_ne: "list_of A \<noteq> []"
  proof (cases "list_of A")
   case Nil
    from multiset_of_list_of[symmetric, of A] and this have "A = {#}" by auto
    from this and A_ne show ?thesis by auto
   next
   case Cons thus ?thesis by auto
  qed
 from assms and multiset_of_list_of
  have P1: "?lt (multiset_of (list_of A))"
  and  P2: "?li (multiset_of (list_of A))" by auto
 from P1 and P2 and local_maxima_multiset_aux and lA_ne
 have "\<exists>a. a \<in># (multiset_of (list_of A)) \<and> (\<forall>b. b \<in># (multiset_of (list_of A)) \<longrightarrow> (b,a) \<notin> s)"
   by metis
 from this and multiset_of_list_of show ?thesis by simp
qed

lemma empty_le[simp]: "{#} \<le> A"
  unfolding mset_le_exists_conv by auto

lemma le_empty[simp]: "(M \<le> {#}) = (M = {#})"
  unfolding mset_le_exists_conv by auto

lemma mset_le_trans:
  assumes "A \<le> B" and "B \<le> C"
  shows "A \<le> (C::_multiset)"
using assms order_trans_rules(23) by simp

lemma set_of_mono:
  assumes le: "A \<le> B"
  shows "set_of A \<subseteq> set_of B" 
proof 
  fix x
  assume "x \<in> set_of A"
  hence "x \<in># A" unfolding mem_set_of_iff by auto
  with mset_leD[OF le] have "x \<in># B" by auto
  thus "x \<in> set_of B" unfolding mem_set_of_iff by auto
qed

abbreviation mset_sum where "mset_sum \<equiv> fold_mset (op +) (0 :: 'a :: comm_monoid_add)"

lemma comp_fun_commute_plus: "comp_fun_commute (op + :: (('a :: comm_monoid_add) \<Rightarrow> 'a \<Rightarrow> 'a))"
  by (unfold_locales, unfold o_def, rule ext, auto simp: ac_simps)

lemma [simp]: "mset_sum (M + N) = mset_sum M + mset_sum N"
  unfolding comp_fun_commute.fold_mset_union[OF comp_fun_commute_plus] 
  unfolding comp_fun_commute.fold_mset_commute[OF comp_fun_commute_plus]
  by simp

lemma [simp]: "mset_sum {# x #} = x"
proof -
  have "mset_sum ({#x#}) = mset_sum ({#} + {#x#})" by simp
  also have "... = x" 
    unfolding comp_fun_commute.fold_mset_insert[OF comp_fun_commute_plus] by simp
  finally show ?thesis .
qed

lemma mult1_ab_a: "({#a#}, {#a,b#}) \<in> mult1 r"
  unfolding mult1_def
proof(rule, unfold split, intro exI conjI)
  show "{#a, b#} = {#a#} + {#b #}"
    by (simp add: ac_simps)
next
  show "{#a#} = {# a #} + {#}" by simp
qed simp

lemma mult1_ab_b: "({#b#}, {#a,b#}) \<in> mult1 r"
proof -
  have id: "{#a,b#} = {#b,a#}" by (simp add: ac_simps)
  show ?thesis unfolding id by (rule mult1_ab_a)
qed

lemma mult1_mono_left[intro!]: assumes "(M,N) \<in> mult1 r"
  shows "((T + M, T + N) \<in> mult1 r)"
proof -
  from assms[unfolded mult1_def]
  obtain a M0 K where id: "N = M0 + {#a#}"
    "M = M0 + K" and rel: "\<forall> b. b \<in># K \<longrightarrow> (b,a) \<in> r" by blast
  show ?thesis unfolding mult1_def
    by (rule, unfold split, rule exI[of _ a], rule exI[of _ "T + M0"], 
      rule exI[of _ K], unfold id, insert rel, auto simp: ac_simps)
qed

lemma mult1_mono_right[intro!]: assumes "(M,N) \<in> mult1 r"
  shows "((M + T, N + T) \<in> mult1 r)"
proof -
  from assms[unfolded mult1_def]
  obtain a M0 K where id: "N = M0 + {#a#}"
    "M = M0 + K" and rel: "\<forall> b. b \<in># K \<longrightarrow> (b,a) \<in> r" by blast
  show ?thesis unfolding mult1_def
    by (rule, unfold split, rule exI[of _ a], rule exI[of _ "T + M0"], 
      rule exI[of _ K], unfold id, insert rel, auto simp: ac_simps)
qed

lemma mult1_singleton[simp]: 
  shows "((M, {# t #}) \<in> mult1 r) = (\<forall> a. a \<in># M \<longrightarrow> (a,t) \<in> r)" (is "?l = ?r")
proof 
  assume r: ?r
  show ?l
    unfolding mult1_def
    by (rule, unfold split, rule exI[of _ t], rule exI[of _ "{#}"], 
      rule exI[of _ M], insert r, auto)
next
  assume l: ?l
  from this[unfolded mult1_def]
  obtain a M0 K where id: "{#t#} = M0 + {#a#}"
    "M = M0 + K" and rel: "\<forall> b. b \<in># K \<longrightarrow> (b,a) \<in> r" by blast
  from id(1)[unfolded single_is_union] have "M0 = {#}" by simp
  with id have id: "t = a" "M = K" by auto
  show ?r unfolding id by (rule rel)
qed

fun ms_lesseq :: "'a list \<Rightarrow> 'a list \<Rightarrow> bool"
  where "ms_lesseq [] _ = True"
 |  "ms_lesseq (Cons x xs) ys = (case extract (op = x) ys of None \<Rightarrow> False
      |  Some (ys1,_,ys2) \<Rightarrow> ms_lesseq xs (ys1 @ ys2))"

lemma ms_lesseq[code_unfold]: "(multiset_of xs \<le> multiset_of ys) = (ms_lesseq xs ys)"
proof (induct xs arbitrary: ys)
  case (Nil ys)
  show ?case by simp
next
  case (Cons x xs ys)
  show ?case 
  proof (cases "extract (op = x) ys")
    case None
    from extract_None[OF this] have x: "x \<notin> set ys" by auto
    {
      assume "multiset_of (x # xs) \<le> multiset_of ys"
      from set_of_mono[OF this] x have False by simp
    }
    with None show ?thesis by auto
  next
    case (Some res)
    obtain ys1 y ys2 where res: "res = (ys1,y,ys2)" by (cases res, auto)
    note Some = Some[unfolded res]
    from extract_Some[OF Some] have ys: "ys = ys1 @ x # ys2" by simp
    show ?thesis unfolding ms_lesseq.simps 
      unfolding Some option.simps split
      unfolding ys
      unfolding Cons[of "ys1 @ ys2", symmetric]
      unfolding mset_le_def by auto
  qed
qed
  
end
