(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Tree_Automata 
imports
  Term_Impl
  Trs
  Monadic_Check
  RBT_Map_Set_Extension
  Transitive_Closure_RBT_Impl
begin


(* syntax similar to AFP-Entry of TA *)
datatype ('q,'f)ta_rule = TA_rule 'f "'q list" 'q ("_ _ \<rightarrow> _")

record ('q,'f)ta = 
  ta_final :: "'q set"
  ta_rules :: "('q,'f) ta_rule set"
  ta_eps   :: "('q \<times> 'q)set"


fun rsym where "rsym (f qs \<rightarrow> q) = (f, length qs)"
fun rqs where "rqs (f qs \<rightarrow> q) = qs"
fun rq where "rq (f qs \<rightarrow> q) = q"


fun ta_res :: "('q,'f)ta \<Rightarrow> ('f,'q)term \<Rightarrow> 'q set"
where "ta_res TA (Var q) = {q'. (q,q') \<in> (ta_eps TA)^*}"
  | "ta_res TA (Fun f ts) = {q' | q' q qs. (f qs \<rightarrow> q) \<in> ta_rules TA \<and> (q,q') \<in> ((ta_eps TA)^*) \<and> length qs = length ts \<and> (\<forall> i < length ts. qs ! i \<in> (map (ta_res TA) ts) ! i)}"

declare ta_res.simps[code del]

definition ta_lang :: "('q,'f)ta \<Rightarrow> ('f,'q)term set"
where "ta_lang TA = {t. ground t \<and> ta_final TA \<inter> ta_res TA t \<noteq> {}}"


lemma ta_res_eps: assumes q: "q \<in> ta_res TA t" and q': "(q,q') \<in> (ta_eps TA)^*"
  shows "q' \<in> ta_res TA t"
  using q
proof (induct t)
  case (Var q'')
  from this[simplified] q' 
  show ?case by simp
next
  case (Fun f ts)
  from Fun(2)
  obtain q'' qs where rule: "f qs \<rightarrow> q'' \<in> ta_rules TA"
    and len: "length qs = length ts"
    and rec: "\<And>i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (ts ! i)"
    and q: "(q'',q) \<in> (ta_eps TA)^*"
    by auto
  from q q' have q': "(q'',q') \<in> (ta_eps TA)^*" by auto
  with rule len rec show ?case by auto
qed
  


(* lemma 32(i) *)
lemma ta_res_subst: assumes 
  x: "\<And> x. x \<in> vars_term t \<Longrightarrow> \<tau> x \<in> ta_res TA (\<sigma> x)"
  and q: "q \<in> ta_res TA (map_vars \<tau> t)"
  shows "q \<in> ta_res TA (t \<cdot> \<sigma>)"
using x q
proof (induct t arbitrary: q)
  case (Var x)
  from Var(1)[of x] Var(2) ta_res_eps
  show ?case by simp
next
  case (Fun f ts)
  from Fun(3)
  obtain q' qs where rule: "f qs \<rightarrow> q' \<in> ta_rules TA"
    and len: "length qs = length ts"
    and rec: "\<And>i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (map_vars \<tau> (ts ! i))"
    and q: "(q',q) \<in> (ta_eps TA)^*"
    by auto
  let ?ts = "map (\<lambda> t. t \<cdot> \<sigma>) ts"
  {
    fix i
    assume i: "i < length ts" 
    from i have mem: "ts ! i \<in> set ts" by auto
    hence "vars_term (ts ! i) \<subseteq> vars_term (Fun f ts)" by auto 
    with Fun(1)[OF mem Fun(2) rec[OF i]] have "qs ! i \<in> ta_res TA (ts ! i \<cdot> \<sigma>)"
      by auto
  } note main = this
  have id: "Fun f ts \<cdot> \<sigma> = Fun f ?ts" by simp
  show ?case
    by (unfold id ta_res.simps, rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ qs], intro conjI, simp, rule rule, rule q, simp add: len, simp add: main)
qed


(* corresponds to lemma 32(ii) *)
lemma ta_res_subst_linear: 
  assumes l: "linear_term t"
  and q: "q \<in> ta_res TA (t \<cdot> \<sigma>)"
  shows "\<exists> \<tau>. q \<in> ta_res TA (map_vars \<tau> t) \<and> (\<forall> x \<in> vars_term t. \<tau> x \<in> ta_res TA (\<sigma> x))"
using l q
proof (induct t arbitrary: q)
  case (Var x)
  from Var(2) show ?case by auto
next
  case (Fun f ts)
  from Fun(3)
  obtain q' qs where rule: "f qs \<rightarrow> q' \<in> ta_rules TA"
    and len: "length qs = length ts"
    and rec: "\<And>i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (ts ! i \<cdot> \<sigma>)"
    and q: "(q',q) \<in> (ta_eps TA)^*"
    by auto
  {
    fix i
    assume i: "i < length ts" 
    from i have mem: "ts ! i \<in> set ts" by auto
    from Fun(2) mem have "linear_term (ts ! i)" by auto
    note Fun(1)[OF mem this rec[OF i]]
  }
  hence "\<forall> i. \<exists> \<tau>. i < length ts \<longrightarrow> qs ! i \<in> ta_res TA (map_vars \<tau> (ts ! i)) \<and>
             (\<forall> x \<in> vars_term (ts ! i). \<tau> x \<in> ta_res TA (\<sigma> x))"
    by blast
  from choice[OF this] obtain \<tau> where
    ind: "\<And> i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (map_vars (\<tau> i) (ts ! i)) \<and> 
             (\<forall> x \<in> vars_term (ts ! i). (\<tau> i) x \<in> ta_res TA (\<sigma> x))"
    by auto
  from Fun(2) have "is_partition (map vars_term ts)" by auto
  from subst_merge[OF this, of \<tau>] obtain \<tau>2 where \<tau>2: "\<And> i. i < length ts \<Longrightarrow> \<forall> x \<in> vars_term (ts ! i). \<tau>2 x = \<tau> i x" by auto
  show ?case
  proof (rule exI[of _ \<tau>2], intro conjI)
    show "\<forall> x \<in> vars_term (Fun f ts). \<tau>2 x \<in> ta_res TA (\<sigma> x)"
    proof
      fix x
      assume "x \<in> vars_term (Fun f ts)"
      then obtain t where "t \<in> set ts" and "x \<in> vars_term t" by auto
      then obtain i where i: "i < length ts" and x: "x \<in> vars_term (ts ! i)" unfolding set_conv_nth by auto
      from ind[OF i] \<tau>2[OF i] x show "\<tau>2 x \<in> ta_res TA (\<sigma> x)" by auto
    qed
  next
    { 
      fix i
      assume i: "i < length ts"
      have id: "map_vars \<tau>2 (ts ! i) = map_vars (\<tau> i) (ts ! i)" 
        by (rule map_vars_vars_term, insert \<tau>2[OF i], auto)
      from ind[OF i, THEN conjunct1] 
      have "qs ! i \<in> ta_res TA (map_vars \<tau>2 (ts ! i))" unfolding id .
    } note main = this
    have "q \<in> ta_res TA (Fun f (map (map_vars \<tau>2) ts))"
      by (unfold ta_res.simps, rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ qs], intro conjI, simp, rule rule, rule q, simp add: len, simp add: main)
    thus "q \<in> ta_res TA (map_vars \<tau>2 (Fun f ts))" by auto
  qed
qed

definition ta_rqs :: "('q,'f)ta \<Rightarrow> 'q set"
where "ta_rqs TA \<equiv> {q'. \<exists> q. (q \<in> rq ` ta_rules TA) \<and> (q,q') \<in> (ta_eps TA)^*}"

lemma ta_rqs_res: assumes ground: "ground t" 
  shows "ta_res TA t \<subseteq> ta_rqs TA"
proof
  fix q 
  assume q: "q \<in> ta_res TA t"
  from ground obtain f ts where t: "t = Fun f ts" by (cases t, auto)
  from q[unfolded t] obtain q' qs where "(f qs \<rightarrow> q') \<in> ta_rules TA" 
    and q: "(q',q) \<in> (ta_eps TA)^*" by auto
  thus "q \<in> ta_rqs TA" unfolding ta_rqs_def by force
qed    


definition ta_syms :: "('q,'f)ta \<Rightarrow> 'f sig"
where "ta_syms TA \<equiv> rsym ` ta_rules TA"

lemma ta_syms_res: "q \<in> ta_res TA t \<Longrightarrow> funas_term t \<subseteq> ta_syms TA"
proof (induct t arbitrary: q)
  case (Var x)
  show ?case by simp
next
  case (Fun f ts)
  from Fun(2)
  obtain q' qs where rule: "f qs \<rightarrow> q' \<in> ta_rules TA"
    and len: "length qs = length ts"
    and rec: "\<forall> i < length ts. qs ! i \<in> ta_res TA (ts ! i)"
    by auto
  {
    fix t 
    assume mem: "t \<in> set ts"
    then obtain i where i: "i < length ts"
      and t: "t = ts ! i" unfolding set_conv_nth by auto
    from Fun(1)[OF mem[unfolded t] rec[THEN spec, THEN mp[OF _ i]]]
    have "funas_term t \<subseteq> ta_syms TA" unfolding t by auto
  } note rec = this
  from rule len have f: "(f,length ts) \<in> ta_syms TA" unfolding ta_syms_def
    by force
  show ?case using f rec by auto
qed

lemma ta_syms_lang: assumes "t \<in> ta_lang TA" shows "funas_term t \<subseteq> ta_syms TA"
proof -
  from assms
  obtain q where "q \<in> ta_res TA t" unfolding ta_lang_def
    by auto
  from ta_syms_res[OF this] show ?thesis  .
qed


(* the restrictions on vars_term and \<tau> are made to ensure that we will only
   have to test finitely many \<tau> *)
definition ta_rule_compatible :: "('q,'f)ta \<Rightarrow> ('f,'v)rule \<Rightarrow> bool"
where "ta_rule_compatible TA \<equiv> \<lambda> (l,r). (\<forall> \<tau>. (\<tau> ` vars_term l \<subseteq> ta_rqs TA) \<longrightarrow> 
  ta_res TA (map_vars \<tau> l) \<subseteq> ta_res TA (map_vars \<tau> r))"


fun ta_match :: "('q,'f)ta \<Rightarrow> ('f,'v)term \<Rightarrow> 'q \<Rightarrow> ('v \<Rightarrow> 'q) set"
  where "ta_match TA (Var x) q = { (\<lambda> _. q') | q'. (q',q) \<in> (ta_eps TA)^*}"
   | "ta_match TA (Fun f ts) q = { fun_merge \<sigma>s (map vars_term ts) | \<sigma>s qs q'. 
          (f qs \<rightarrow> q') \<in> ta_rules TA \<and> 
          (q',q) \<in> (ta_eps TA)^* \<and> 
          length qs = length ts \<and>
          length \<sigma>s = length ts \<and>
          (\<forall> i < length ts. \<sigma>s ! i \<in> ta_match TA (ts ! i) (qs ! i)) }"


lemma ta_match: 
  assumes "q \<in> ta_res TA (map_vars \<tau> t)"
                shows "\<exists> \<sigma> \<in> ta_match TA t q. (\<forall> x \<in> vars_term t. \<tau> x = \<sigma> x)"
  using assms
proof (induct t arbitrary: q)
  case (Var x)
  thus ?case by auto
next
  case (Fun f ts)
  from Fun(2)
  obtain q' qs where rule: "f qs \<rightarrow> q' \<in> ta_rules TA"
    and eps: "(q',q) \<in> (ta_eps TA)^*"
    and len: "length qs = length ts"
    and rec: "\<And> i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (map_vars \<tau> (ts ! i))"
    by auto
  let ?p = "\<lambda> i \<sigma>. \<sigma> \<in> ta_match TA (ts ! i) (qs ! i) \<and> (\<forall> x \<in> vars_term (ts ! i). \<tau> x = \<sigma> x)" 
  {
    fix i
    assume i: "i < length ts"
    hence ti: "ts ! i \<in> set ts" by auto
    from Fun(1)[OF ti rec[OF i]] 
    have "\<exists> \<sigma>. ?p i \<sigma>" by blast
  }
  hence "\<forall> i. \<exists> \<sigma>. i < length ts \<longrightarrow> ?p i \<sigma>" by auto
  from choice[OF this] obtain \<sigma>s where \<sigma>s: "\<And> i. i < length ts \<Longrightarrow> ?p i (\<sigma>s i)" by auto
  let ?\<sigma>s = "map \<sigma>s [0 ..< length ts]"
  let ?xs = "map vars_term ts"
  obtain \<sigma> where \<sigma>: "\<sigma> \<equiv> fun_merge ?\<sigma>s ?xs" by auto
  have mem: "\<sigma> \<in> ta_match TA (Fun f ts) q"
    unfolding ta_match.simps
    by (rule, intro exI conjI, unfold \<sigma>, 
    rule HOL.refl,
    rule rule,
    insert len eps \<sigma>s, auto)
  show ?case
  proof (rule bexI[OF _ mem], intro ballI)
    fix x 
    assume "x \<in> vars_term (Fun f ts)"
    then obtain t where "t \<in> set ts" and x: "x \<in> vars_term t" by auto
    then obtain i where i: "i < length ts" and t: "t = ts ! i" 
      unfolding set_conv_nth by auto
    from i x[unfolded t] have x: "x \<in> (map vars_term ts) ! i" by auto
    (* from Fun(2) have part: "is_partition (map vars_term ts)" by simp *)
    have "\<sigma> x = (?\<sigma>s ! i) x" unfolding \<sigma> 
    proof (rule fun_merge[OF _ x, of ?\<sigma>s, unfolded length_map, OF i])
      fix i j x
      assume i: "i < length ts" and j: "j < length ts" and xi: "x \<in> ?xs ! i" and xj: "x \<in> ?xs ! j"
      have "(?\<sigma>s ! i) x = \<tau> x" using \<sigma>s[OF i] xi i by auto 
      also have "... = (?\<sigma>s ! j) x" using \<sigma>s[OF j] xj j by auto 
      finally show "(?\<sigma>s ! i) x = (?\<sigma>s ! j) x" .
    qed
    also have "... = \<tau> x"
      using \<sigma>s[OF i] x i by auto
    finally show "\<tau> x = \<sigma> x" by simp
  qed
qed


(* TODO: examine whether \<sigma> in ta_match is always able to produce q,
   then one can ignore the precondition "q \<in> ta_res TA (Fun f ls \<cdot> \<sigma>)" *)
fun ta_rule_compatible_eff :: "('q,'f)ta \<Rightarrow> ('f,'v)rule \<Rightarrow> bool"
where "ta_rule_compatible_eff TA (Fun f ls,r) = (\<forall> \<sigma> qs q.
  f qs \<rightarrow> q \<in> ta_rules TA \<and>
  length qs = length ls \<and>
  \<sigma> \<in> ta_match TA (Fun f ls) q \<and>
  q \<in> ta_res TA (map_vars \<sigma> (Fun f ls)) \<longrightarrow>
  q \<in> ta_res TA (map_vars \<sigma> r))"
 | "ta_rule_compatible_eff TA (Var x,r) = (\<forall> q \<in> ta_rqs TA. q \<in> ta_res TA (map_vars (\<lambda> _.q) r))"


lemma ta_rule_compatible_eff: 
  assumes var: "vars_term l \<supseteq> vars_term r"
  and compat: "ta_rule_compatible_eff TA (l,r)"
  shows "ta_rule_compatible TA (l,r)"
  unfolding ta_rule_compatible_def split
proof (intro allI impI)
  fix \<tau>
  assume \<tau>: "\<tau> ` vars_term l \<subseteq> ta_rqs TA"
  show "ta_res TA (map_vars \<tau> l) \<subseteq> ta_res TA (map_vars \<tau> r)" (is "?rl \<subseteq> ?rr")
  proof (cases l)
    case (Fun f ls) note l = this
    show ?thesis
    proof
      fix q'
      assume q': "q' \<in> ?rl"
      then obtain qs q where rule: "f qs \<rightarrow> q \<in> ta_rules TA" and len: "length qs = length ls" 
        and eps: "(q,q') \<in> (ta_eps TA)^*" and q: "q \<in> ?rl" unfolding l
        by auto
      from ta_match[OF q] obtain \<sigma> where \<sigma>: "\<sigma> \<in> ta_match TA l q" and \<tau>: "\<And> x. x \<in> vars_term l \<Longrightarrow> \<tau> x = \<sigma> x" by auto
      have id: "map_vars \<tau> l = map_vars \<sigma> l"
        by (rule map_vars_vars_term, insert \<tau>, auto)
      from compat[unfolded l ta_rule_compatible_eff.simps, THEN spec[of _ \<sigma>], THEN spec[of _ qs], THEN spec[of _ q]] rule len \<sigma> l q[unfolded id] have q: "q \<in> ta_res TA (map_vars \<sigma> r)" by simp
      have "map_vars \<sigma> r = map_vars \<tau> r"
      proof (rule map_vars_vars_term)
        fix x
        assume "x \<in> vars_term r"
        with var have "x \<in> vars_term l" by auto
        with \<tau> show "\<sigma> x = \<tau> x" by simp
      qed
      with q have q: "q \<in> ta_res TA (map_vars \<tau> r)" by simp
      from ta_res_eps[OF q eps] show "q' \<in> ta_res TA (map_vars \<tau> r)"  .
    qed
  next
    case (Var x)
    with \<tau> have \<tau>: "\<tau> x \<in> ta_rqs TA" by auto
    let ?\<tau> = "\<lambda> _ . \<tau> x"
    from compat[unfolded Var ta_rule_compatible_eff.simps, THEN bspec[OF _ \<tau>]]
    have \<tau>r: "\<tau> x \<in> ta_res TA (map_vars ?\<tau> r)" by simp
    have "map_vars ?\<tau> r = map_vars \<tau> r"
    proof (rule map_vars_vars_term)
      fix y
      assume "y \<in> vars_term r"
      with var have "y \<in> vars_term l" by auto
      with Var have "y = x" by simp
      thus "\<tau> x = \<tau> y" by simp
    qed
    with \<tau>r have \<tau>r: "\<tau> x \<in> ta_res TA (map_vars \<tau> r)" by simp
    {
      fix q'
      assume "q' \<in> ?rl" 
      hence "(\<tau> x,q') \<in> (ta_eps TA)^*" unfolding Var by simp
      from ta_res_eps[OF \<tau>r this] have "q' \<in> ?rr" .
    }        
    thus ?thesis by auto
  qed
qed

definition ta_trs_compatible :: "('q,'f)ta \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
where "ta_trs_compatible TA R \<equiv> \<forall> l r. (l,r) \<in> R \<and> funas_term l \<subseteq> ta_syms TA \<longrightarrow> ta_rule_compatible TA (l,r)" 


lemma ta_trs_compatible_res: assumes 
      comp: "ta_trs_compatible TA R"
  and ll: "left_linear_trs R"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and step: "(s,t) \<in> rstep R"
  and ground: "ground s"
  shows "ta_res TA s \<subseteq> ta_res TA t"
  using step ground
proof (induct)
  case (IH C \<sigma> l r)
  hence ground: "ground (l \<cdot> \<sigma>)" by simp
  show ?case
  proof (induct C)
    case (More f bef C aft)
    let ?args = "\<lambda> l. bef @ C\<langle>l \<cdot> \<sigma>\<rangle> # aft"
    let ?l = "Fun f (?args l)"
    let ?r = "Fun f (?args r)"
    let ?n = "Suc (length bef + length aft)"
    let ?m = "length bef"
    show ?case
    proof
      fix q
      assume "q \<in> ta_res TA (More f bef C aft)\<langle>l\<cdot>\<sigma>\<rangle>"
      hence "q \<in> ta_res TA ?l" by simp
      from this[unfolded ta_res.simps] obtain q' qs where 
        rule: "f qs \<rightarrow> q' \<in> ta_rules TA"
        and len: "length qs = ?n"
        and rec: "\<forall>i < ?n. qs ! i \<in> map (ta_res TA) (?args l) ! i"
        and q: "(q',q) \<in> (ta_eps TA)^*"
        by auto
      have len2: "\<And>t. ?m < length (?args t)" by auto
      from rec[THEN spec, of ?m, unfolded nth_map[OF len2]]
      have "qs ! ?m \<in> ta_res TA (C\<langle>l \<cdot> \<sigma>\<rangle>)"  by auto
      with More have m: "qs ! ?m \<in> ta_res TA (C\<langle>r\<cdot>\<sigma>\<rangle>)" by auto
      { 
        fix i
        assume i: "i < ?n"
        have "qs ! i \<in> map (ta_res TA) (?args r) ! i" 
        proof (cases "i = ?m")
          case True
          show ?thesis using m unfolding True
            by (simp add: nth_append)
        next
          case False
          note rec = rec[THEN spec, THEN mp, OF i]
          show ?thesis 
          proof (cases "i < ?m")
            case True
            show ?thesis using rec
              by (simp add: nth_append True)
          next
            case False
            with `i \<noteq> ?m` have "i - ?m > 0" by auto
            then obtain j where im: "i - ?m = Suc j" by (cases "i - ?m", auto)
            show ?thesis using rec 
              by (simp add: nth_append False im)
          qed
        qed
      }
      hence rec: "\<forall>i < ?n. qs ! i \<in> map (ta_res TA) (?args r) ! i"
        by simp        
      have "q \<in> ta_res TA ?r" 
        unfolding ta_res.simps
        by (rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ qs],
          insert rule len rec q, simp)
      thus "q \<in> ta_res TA (More f bef C aft)\<langle>r\<cdot>\<sigma>\<rangle>" by simp
    qed
  next
    case Hole
    show ?case
    proof (unfold ctxt_apply.simps, rule)
      fix q
      assume q: "q \<in> ta_res TA (l \<cdot> \<sigma>)"
      from ll IH(1) have lin: "linear_term l" unfolding left_linear_trs_def by auto
      from ta_res_subst_linear[OF this q] obtain \<tau> where
        q: "q \<in> ta_res TA (map_vars \<tau> l)" and 
        vars: "\<forall> x \<in> vars_term l. \<tau> x \<in> ta_res TA (\<sigma> x)" by auto
      from ta_syms_res[OF q]
      have "funas_term l \<subseteq> ta_syms TA" by auto
      with comp[unfolded ta_trs_compatible_def] IH(1) have
        comp: "ta_rule_compatible TA (l,r)" by auto
      have "\<tau> ` vars_term l \<subseteq> ta_rqs TA" 
      proof 
        fix q
        assume "q \<in> \<tau> ` vars_term l"
        then obtain x where x: "x \<in> vars_term l" and q: "q = \<tau> x" by auto
        with vars have q: "q \<in> ta_res TA (\<sigma> x)" by auto
        from ground x have "ground (\<sigma> x)" by auto
        from ta_rqs_res[OF this] q show "q \<in> ta_rqs TA" by auto
      qed
      with comp[unfolded ta_rule_compatible_def] q
      have q: "q \<in> ta_res TA (map_vars \<tau> r)" by auto
      from ta_res_subst[OF _ q] wf_trs[OF IH(1)] vars
      show "q \<in> ta_res TA (r \<cdot> \<sigma>)" by force
    qed
  qed
qed


lemma ta_trs_compatible_lang: assumes 
      comp: "ta_trs_compatible TA R"
  and ll: "left_linear_trs R"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and step: "(s,t) \<in> rstep R"
  and s: "s \<in> ta_lang TA"
  shows "t \<in> ta_lang TA"
proof -
  from s[unfolded ta_lang_def]
  have g: "ground s" and acc: "ta_final TA \<inter> ta_res TA s \<noteq> {}" by auto
  from rstep_ground[OF wf_trs g step] have gt: "ground t" .
  from ta_trs_compatible_res[OF comp ll wf_trs step g] acc 
  have acc: "ta_final TA \<inter> ta_res TA t \<noteq> {}" by auto
  from gt acc show "t \<in> ta_lang TA" unfolding ta_lang_def by auto
qed


(* lemma 33 *)
lemma ta_trs_compatible: assumes 
      comp: "ta_trs_compatible TA R"
  and ll: "left_linear_trs R"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and L: "L \<subseteq> ta_lang TA"
  shows "{t | s t. s \<in> L \<and> (s,t) \<in> (rstep R)^*} \<subseteq> ta_lang TA"
proof -
  {
    fix s t
    assume "(s,t) \<in> (rstep R)^*" and "s \<in> L" 
    hence "t \<in> ta_lang TA"
    proof (induct)
      case base
      thus ?case using L by auto
    next
      case (step t u)
      hence t: "t \<in> ta_lang TA" by simp
      show ?case
        by (rule ta_trs_compatible_lang[OF comp ll wf_trs step(2) t])
    qed
  }
  thus ?thesis by auto
qed

lemma finite_ta_syms: assumes "finite (ta_rules TA)" shows "finite (ta_syms TA)"
  unfolding ta_syms_def using assms by auto

    

(* lemma 35 *)
lemma ta_compact: assumes finite: "finite (ta_rules TA)"
  and comp: "ta_trs_compatible TA R"
  and ll: "left_linear_trs R"
  and wf_trs: "\<And> l r. (l,r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and L: "L \<subseteq> ta_lang TA"
  shows "compact L R"
  unfolding compact_def
  by (rule exI[of _ "ta_syms TA"], intro conjI[OF finite_ta_syms[OF finite]],
    insert ta_trs_compatible[OF comp ll wf_trs L] ta_syms_lang, auto)

definition ta_contains :: "('f \<times> nat)set \<Rightarrow> ('q,'f)ta \<Rightarrow> 'q set \<Rightarrow> bool"
  where "ta_contains F TA Q \<equiv> Q \<subseteq> ta_final TA
  \<and> (\<forall> f qs. (f,length qs) \<in> F \<and> set qs \<subseteq> Q \<longrightarrow> (\<exists> q q'. (f qs \<rightarrow> q) \<in> ta_rules TA \<and> q' \<in> Q \<and> (q,q') \<in> (ta_eps TA)^*))"

lemma ta_contains: 
  assumes contain: "ta_contains F TA qs"
  shows "({t :: ('f,'q)term. funas_term t \<subseteq> F}) \<inter> {t. ground t} \<subseteq> ta_lang TA"
proof (clarify)
  fix t :: "('f,'q)term"
  note contain = contain[unfolded ta_contains_def, simplified]
  assume "funas_term t \<subseteq> F" and g: "ground t"
  hence "\<exists> q \<in> qs. q \<in> ta_res TA t"
  proof (induct t)
    case (Var x) thus ?case by simp
  next
    case (Fun f ts)
    {
      fix i
      assume "i < length ts"
      hence t: "ts ! i \<in> set ts" by auto
      from Fun(3) t have gt: "ground (ts ! i)" by auto
      from Fun(2) t have fs: "funas_term (ts ! i) \<subseteq> F" by auto
      note Fun(1)[OF t fs gt] 
    } hence "\<forall> i. \<exists> q. i < length ts \<longrightarrow> q \<in> qs \<and> q \<in> ta_res TA (ts ! i)" by auto
    from choice[OF this] obtain qs' where qs'res: "\<And>i. i < length ts \<Longrightarrow> qs' i \<in> qs \<and> qs' i \<in> ta_res TA (ts ! i)" by auto
    let ?qs = "map qs' [0 ..< length ts]"
    have qs': "\<And> i. i < length ts \<Longrightarrow>  ?qs ! i = qs' i" by auto
    with qs'res have qs'res: "\<forall> i < length ts. ?qs ! i \<in> qs \<and> ?qs ! i \<in> ta_res TA (ts ! i)" by auto
    from contain[THEN conjunct2, THEN spec[of _ f], THEN spec[of _ ?qs]] Fun(2) qs'res 
    obtain q' q where q: "q \<in> qs" and q': "(q',q) \<in> (ta_eps TA)^*" and rule: "f ?qs \<rightarrow> q' \<in> ta_rules TA" by force
    show ?case unfolding ta_res.simps
      by (rule bexI[OF _ q], rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ ?qs], insert rule qs'res, simp add: q')
  qed
  then obtain q where q1: "q \<in> qs" and q2: "q \<in> ta_res TA t" by auto
  from contain[THEN conjunct1] q1 have q1: "q \<in> ta_final TA"
    by auto
  show "t \<in> ta_lang TA" unfolding ta_lang_def using q1 q2 g by auto
qed  


end
